@@ -46,20 +46,21 @@ get_current_time (gchar time_string[100])
 		strftime (time_string, 100, "%Y-%m-%dT%H:%M:%SZ", tm);
 }
 
-static char *
-vcard_from_couch_document (CouchdbDocument *document)
+static EContact *
+contact_from_couch_document (CouchdbDocument *document)
 {
 	EContact *contact;
 	char *str;
 	GSList *list, *sl;
 	GList *attr_list;
 	CouchdbStructField *app_annotations;
+	EContactName contact_name;
 
-	if (!desktopcouch_document_is_contact (document))
+	if (!desktopcouch_document_is_contact (DESKTOPCOUCH_DOCUMENT (document)))
 		return NULL;
 
 	/* Check if the contact is marked for deletion */
-	if ((app_annotations = desktopcouch_document_get_application_annotations (document))) {
+	if ((app_annotations = desktopcouch_document_get_application_annotations (DESKTOPCOUCH_DOCUMENT (document)))) {
 		CouchdbStructField *u1_annotations;
 
 		u1_annotations = couchdb_struct_field_get_struct_field (
@@ -72,7 +73,7 @@ vcard_from_couch_document (CouchdbDocument *document)
 			if (private_annotations != NULL) {
 				if (couchdb_struct_field_has_field (private_annotations, "deleted")
 				    && couchdb_struct_field_get_boolean_field (private_annotations, "deleted"))
-					couchdb_struct_field_unref (app_annotations);
+					g_object_unref (app_annotations);
 					return NULL;
 			}
 		}
@@ -85,38 +86,44 @@ vcard_from_couch_document (CouchdbDocument *document)
 					  couchdb_document_get_revision (document));
 
 	e_contact_set (contact, E_CONTACT_UID, (const gpointer) couchdb_document_get_id (document));
-	e_contact_set (contact, E_CONTACT_GIVEN_NAME,
-		       (const gpointer) desktopcouch_document_contact_get_first_name (document));
-	e_contact_set (contact, E_CONTACT_FAMILY_NAME,
-		       (const gpointer) desktopcouch_document_contact_get_last_name (document));
+
+	contact_name.family = desktopcouch_document_contact_get_last_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
+	contact_name.given = desktopcouch_document_contact_get_first_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
+	contact_name.additional = desktopcouch_document_contact_get_middle_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
+	contact_name.prefixes = desktopcouch_document_contact_get_title (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
+	contact_name.suffixes = desktopcouch_document_contact_get_suffix (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
+	e_contact_set (contact, E_CONTACT_NAME, (const gpointer) &contact_name);
+
+	str = e_contact_name_to_string (&contact_name);
+	e_contact_set (contact, E_CONTACT_FULL_NAME, (const gpointer) str);
+	g_free (str);
+
 	e_contact_set (contact, E_CONTACT_NICKNAME,
-		       (const gpointer) desktopcouch_document_contact_get_nick_name (document));
+		       (const gpointer) desktopcouch_document_contact_get_nick_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_SPOUSE,
-		       (const gpointer) desktopcouch_document_contact_get_spouse_name (document));
+		       (const gpointer) desktopcouch_document_contact_get_spouse_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 
 	e_contact_set (contact, E_CONTACT_ORG,
-		       (const gpointer) desktopcouch_document_contact_get_company (document));
+		       (const gpointer) desktopcouch_document_contact_get_company (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_ORG_UNIT,
-		       (const gpointer) desktopcouch_document_contact_get_department (document));
+		       (const gpointer) desktopcouch_document_contact_get_department (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_TITLE,
-		       (const gpointer) desktopcouch_document_contact_get_title (document));
-	e_contact_set (contact, E_CONTACT_ROLE,
-		       (const gpointer) desktopcouch_document_contact_get_job_title (document));
+		       (const gpointer) desktopcouch_document_contact_get_job_title (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_MANAGER,
-		       (const gpointer) desktopcouch_document_contact_get_manager_name (document));
+		       (const gpointer) desktopcouch_document_contact_get_manager_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_ASSISTANT,
-		       (const gpointer) desktopcouch_document_contact_get_assistant_name (document));
+		       (const gpointer) desktopcouch_document_contact_get_assistant_name (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_OFFICE,
-		       (const gpointer) desktopcouch_document_contact_get_office (document));
+		       (const gpointer) desktopcouch_document_contact_get_office (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_CATEGORIES,
-		       (const gpointer) desktopcouch_document_contact_get_categories (document));
+		       (const gpointer) desktopcouch_document_contact_get_categories (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 	e_contact_set (contact, E_CONTACT_NOTE,
-		       (const gpointer) desktopcouch_document_contact_get_notes (document));
+		       (const gpointer) desktopcouch_document_contact_get_notes (DESKTOPCOUCH_DOCUMENT_CONTACT (document)));
 
 	/* parse email addresses */
 	attr_list = NULL;
 
-	list = desktopcouch_document_contact_get_email_addresses (document);
+	list = desktopcouch_document_contact_get_email_addresses (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	while (list != NULL) {
 		const char *email_str, *description_str, *uuid_str;
 		EVCardAttribute *attr;
@@ -152,7 +159,7 @@ vcard_from_couch_document (CouchdbDocument *document)
 
 		/* remove address from list */
 		list = g_slist_remove (list, email_address);
-		couchdb_struct_field_unref (email_address);
+		g_object_unref (email_address);
 	}
 
 	if (attr_list) {
@@ -162,7 +169,7 @@ vcard_from_couch_document (CouchdbDocument *document)
 	}
 
 	/* parse phone numbers */
-	list = desktopcouch_document_contact_get_phone_numbers (document);
+	list = desktopcouch_document_contact_get_phone_numbers (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	while (list != NULL) {
 		const char *phone_str, *description_str, *uuid_str;
 		EVCardAttribute *attr;
@@ -270,13 +277,12 @@ vcard_from_couch_document (CouchdbDocument *document)
 
 		/* remove phones from list */
 		list = g_slist_remove (list, phone_number);
-		couchdb_struct_field_unref (phone_number);
+		g_object_unref (phone_number);
 	}
 
 	/* parse postal addresses */
-	list = desktopcouch_document_contact_get_addresses (document);
+	list = desktopcouch_document_contact_get_addresses (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	while (list != NULL) {
-		char **street_lines;
 		const char *description_str;
 		EContactAddress *contact_address;
 		CouchdbStructField *address = (CouchdbStructField *) list->data;
@@ -284,15 +290,8 @@ vcard_from_couch_document (CouchdbDocument *document)
 		contact_address = g_new0 (EContactAddress, 1);
 		contact_address->address_format = g_strdup ("");
 
-		street_lines = g_strsplit (desktopcouch_document_contact_address_get_street (address), "\n", 2);
-		if (street_lines != NULL) {
-			contact_address->street = g_strdup (street_lines[0]);
-			if (street_lines[1] != NULL)
-				contact_address->ext = g_strdup (street_lines[1]);
-			g_strfreev (street_lines);
-		} else
-			contact_address->street = g_strdup (desktopcouch_document_contact_address_get_street (address));
-
+		contact_address->street = g_strdup (desktopcouch_document_contact_address_get_street (address));
+		contact_address->ext = g_strdup (desktopcouch_document_contact_address_get_ext_street (address));
 		contact_address->locality = g_strdup (desktopcouch_document_contact_address_get_city (address));
 		contact_address->region = g_strdup (desktopcouch_document_contact_address_get_state (address));
 		contact_address->country = g_strdup (desktopcouch_document_contact_address_get_country (address));
@@ -309,13 +308,13 @@ vcard_from_couch_document (CouchdbDocument *document)
 
 		/* remove addresses from list */
 		list = g_slist_remove (list, address);
-		couchdb_struct_field_unref (address);
+		g_object_unref (address);
 
 		e_contact_address_free (contact_address);
 	}
 
 	/* parse URLs */
-	list = desktopcouch_document_contact_get_urls (document);
+	list = desktopcouch_document_contact_get_urls (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	while (list != NULL) {
 		const char *description_str, *address_str, *uuid_str;
 		EVCardAttribute *attr;
@@ -345,11 +344,11 @@ vcard_from_couch_document (CouchdbDocument *document)
 
 		/* remove urls from list */
 		list = g_slist_remove (list, url);
-		couchdb_struct_field_unref (url);
+		g_object_unref (url);
 	}
 
 	/* parse IM addresses */
-	list = desktopcouch_document_contact_get_im_addresses (document);
+	list = desktopcouch_document_contact_get_im_addresses (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	while (list != NULL) {
 		const char *address_str, *description_str, *protocol_str, *uuid_str;
 		EVCardAttribute *attr = NULL;
@@ -410,11 +409,11 @@ vcard_from_couch_document (CouchdbDocument *document)
 
 		/* remove addresses from list */
 		list = g_slist_remove (list, im);
-		couchdb_struct_field_unref (im);
+		g_object_unref (im);
 	}
 
 	/* birth date */
-	str = (char *) desktopcouch_document_contact_get_birth_date (document);
+	str = (char *) desktopcouch_document_contact_get_birth_date (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	if (str && strlen (str) > 0) {
 		EContactDate *dt;
 
@@ -426,7 +425,7 @@ vcard_from_couch_document (CouchdbDocument *document)
 	}
 
 	/* wedding date */
-	str = (char *) desktopcouch_document_contact_get_wedding_date (document);
+	str = (char *) desktopcouch_document_contact_get_wedding_date (DESKTOPCOUCH_DOCUMENT_CONTACT (document));
 	if (str && strlen (str)) {
 		EContactDate *dt;
 
@@ -464,7 +463,7 @@ vcard_from_couch_document (CouchdbDocument *document)
 						  str);
 
 		g_free (str);
-		couchdb_struct_field_unref (app_annotations);
+		g_object_unref (app_annotations);
 	} else {
 		/* Always have a REV field on the VCARD, for SyncEvolution (bug LP:#479110) */
 		gchar time_string[100] = {0};
@@ -473,12 +472,7 @@ vcard_from_couch_document (CouchdbDocument *document)
 		e_contact_set (contact, E_CONTACT_REV, time_string);
 	}
 
-	/* convert the contact to a VCARD string to be returned */
-	str = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);
-
-	g_object_unref (G_OBJECT (contact));
-
-	return str;
+	return contact;
 }
 
 static CouchdbStructField *
@@ -600,24 +594,17 @@ contact_postal_address_to_struct_field (EContact *contact, EContactField field,
 
 	contact_address = e_contact_get (contact, field);
 	if (contact_address) {
-		char *street = NULL;
-
-		if (contact_address->ext && *contact_address->ext)
-			street = g_strdup_printf ("%s\n%s", contact_address->street, contact_address->ext);
-		else if (contact_address->street && *contact_address->street)
-			street = g_strdup (contact_address->street);
-
 		sf = desktopcouch_document_contact_address_new (/* FIXME */ NULL,
-							   street,
-							   contact_address->locality,
-							   contact_address->region,
-							   contact_address->country,
-							   contact_address->code,
-							   contact_address->po,
-							   description);
+								contact_address->street,
+								contact_address->ext,
+								contact_address->locality,
+								contact_address->region,
+								contact_address->country,
+								contact_address->code,
+								contact_address->po,
+								description);
 		
 		e_contact_address_free (contact_address);
-		g_free (street);
 	}
 
 	return sf;
@@ -711,7 +698,7 @@ set_vcard_revision (CouchdbStructField *app_annotations, EContact *contact)
 	}
 
 	couchdb_struct_field_set_struct_field (app_annotations, "Evolution", evo_annotations);
-	couchdb_struct_field_unref (evo_annotations);
+	g_object_unref (evo_annotations);
 }
 
 static CouchdbDocument *
@@ -721,30 +708,44 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	GSList *list;
 	GList *attr_list, *al;
 	const char *str;
-	CouchdbDocument *document;
+	DesktopcouchDocumentContact *document;
 	gint i;
 	CouchdbStructField *postal_address, *app_annotations;
+	EContactName *contact_name;
 
 	/* create the CouchDBDocument to put on the database */
-	document = desktopcouch_document_contact_new (couchdb_backend->couchdb);
+	document = desktopcouch_document_contact_new ();
 
 	str = e_contact_get_const (contact, E_CONTACT_UID);
 	if (str)
-		couchdb_document_set_id (document, str);
+		couchdb_document_set_id (COUCHDB_DOCUMENT (document), str);
 
 	str = e_vcard_attribute_get_value (e_vcard_get_attribute (E_VCARD (contact), COUCHDB_REVISION_PROP));
 	if (str)
-		couchdb_document_set_revision (document, str);
+		couchdb_document_set_revision (COUCHDB_DOCUMENT (document), str);
+
+	contact_name = (EContactName *) e_contact_get (contact, E_CONTACT_NAME);
+	if (contact_name != NULL) {
+		if (contact_name->prefixes != NULL)
+			desktopcouch_document_contact_set_title (document, (const char *) contact_name->prefixes);
+		if (contact_name->given != NULL)
+			desktopcouch_document_contact_set_first_name (document, (const char *) contact_name->given);
+		if (contact_name->additional != NULL)
+			desktopcouch_document_contact_set_middle_name (document, (const gchar *) contact_name->additional);
+		if (contact_name->family != NULL)
+			desktopcouch_document_contact_set_last_name (document, (const char *) contact_name->family);
+		if (contact_name->suffixes != NULL)
+			desktopcouch_document_contact_set_suffix (document, (const char *) contact_name->suffixes);
+	}
+
+	e_contact_name_free (contact_name);
 
-	desktopcouch_document_contact_set_first_name (document, (const char *) e_contact_get_const (contact, E_CONTACT_GIVEN_NAME));
-	desktopcouch_document_contact_set_last_name (document, (const char *) e_contact_get_const (contact, E_CONTACT_FAMILY_NAME));
 	desktopcouch_document_contact_set_nick_name (document, (const char *) e_contact_get_const (contact, E_CONTACT_NICKNAME));
 	desktopcouch_document_contact_set_spouse_name (document, (const char *) e_contact_get_const (contact, E_CONTACT_SPOUSE));
 
 	desktopcouch_document_contact_set_company (document, (const char *) e_contact_get_const (contact, E_CONTACT_ORG));
 	desktopcouch_document_contact_set_department (document, (const char *) e_contact_get_const (contact, E_CONTACT_ORG_UNIT));
-	desktopcouch_document_contact_set_title (document, (const char *) e_contact_get_const (contact, E_CONTACT_TITLE));
-	desktopcouch_document_contact_set_job_title (document, (const char *) e_contact_get_const (contact, E_CONTACT_ROLE));
+	desktopcouch_document_contact_set_job_title (document, (const char *) e_contact_get_const (contact, E_CONTACT_TITLE));
 	desktopcouch_document_contact_set_manager_name (document, (const char *) e_contact_get_const (contact, E_CONTACT_MANAGER));
 	desktopcouch_document_contact_set_assistant_name (document, (const char *) e_contact_get_const (contact, E_CONTACT_ASSISTANT));
 	desktopcouch_document_contact_set_office (document, (const char *) e_contact_get_const (contact, E_CONTACT_OFFICE));
@@ -754,20 +755,25 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	/* email addresses */
 	list = NULL;
 	attr_list = e_contact_get_attributes (contact, E_CONTACT_EMAIL);
-	for (al = attr_list; al != NULL; al = al->next) {
-		CouchdbStructField *sf;
-		EVCardAttribute *attr = (EVCardAttribute *) al->data;
+	if (attr_list != NULL) {
+		for (al = attr_list; al != NULL; al = al->next) {
+			CouchdbStructField *sf;
+			EVCardAttribute *attr = (EVCardAttribute *) al->data;
 
-		sf = contact_email_to_struct_field (attr);
-		if (sf)
-			list = g_slist_append (list, sf);
-	}
+			sf = contact_email_to_struct_field (attr);
+			if (sf)
+				list = g_slist_append (list, sf);
+		}
 
-	if (list) {
-		desktopcouch_document_contact_set_email_addresses (document, list);
+		if (list) {
+			desktopcouch_document_contact_set_email_addresses (document, list);
 
-		g_slist_foreach (list, (GFunc) couchdb_struct_field_unref, NULL);
-		g_slist_free (list);
+			g_slist_foreach (list, (GFunc) g_object_unref, NULL);
+			g_slist_free (list);
+		}
+
+		g_list_foreach (attr_list, (GFunc) e_vcard_attribute_free, NULL);
+		g_list_free (attr_list);
 	}
 
 	/* phone numbers */
@@ -787,7 +793,7 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	if (list) {
 		desktopcouch_document_contact_set_phone_numbers (document, list);
 
-		g_slist_foreach (list, (GFunc) couchdb_struct_field_unref, NULL);
+		g_slist_foreach (list, (GFunc) g_object_unref, NULL);
 		g_slist_free (list);
 	}
 
@@ -808,7 +814,7 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	if (list) {
 		desktopcouch_document_contact_set_addresses (document, list);
 
-		g_slist_foreach (list, (GFunc) couchdb_struct_field_unref, NULL);
+		g_slist_foreach (list, (GFunc) g_object_unref, NULL);
 		g_slist_free (list);
 	}
 
@@ -833,7 +839,7 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	if (list != NULL) {
 		desktopcouch_document_contact_set_urls (document, list);
 
-		g_slist_foreach (list, (GFunc) couchdb_struct_field_unref, NULL);
+		g_slist_foreach (list, (GFunc) g_object_unref, NULL);
 		g_slist_free (list);
 	}
 
@@ -876,7 +882,7 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	if (list != NULL) {
 		desktopcouch_document_contact_set_im_addresses (document, list);
 
-		g_slist_foreach (list, (GFunc) couchdb_struct_field_unref, NULL);
+		g_slist_foreach (list, (GFunc) g_object_unref, NULL);
 		g_slist_free (list);
 	}
 
@@ -910,44 +916,39 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 		app_annotations = couchdb_struct_field_new ();
 
 	set_vcard_revision (app_annotations, contact);
-	desktopcouch_document_set_application_annotations (document, app_annotations);
-	couchdb_struct_field_unref (app_annotations);
+	desktopcouch_document_set_application_annotations (DESKTOPCOUCH_DOCUMENT (document), app_annotations);
+	g_object_unref (app_annotations);
 
-	return document;
+	return COUCHDB_DOCUMENT (document);
 }
 
 static void
-document_updated_cb (CouchdbSession *couchdb, const char *dbname, CouchdbDocument *document, gpointer user_data)
+document_updated_cb (CouchdbDatabase *database, CouchdbDocument *document, gpointer user_data)
 {
-	char *vcard;
 	EContact *contact;
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (user_data);
 
-	if (g_strcmp0 (dbname, couchdb_backend->dbname) != 0)
+	contact = contact_from_couch_document (document);
+	if (!contact)
 		return;
 
-	vcard = vcard_from_couch_document (document);
-	if (!vcard)
-		return;
+	e_book_backend_notify_update (E_BOOK_BACKEND (couchdb_backend), contact);
 
-	contact = e_contact_new_from_vcard (vcard);
-	if (contact != NULL) {
-		e_book_backend_notify_update (E_BOOK_BACKEND (couchdb_backend), contact);
-		g_object_unref (G_OBJECT (contact));
-	}
+	/* Add the contact to the cache */
+	e_book_backend_cache_add_contact (couchdb_backend->cache, contact);
 
-	g_free (vcard);
+	g_object_unref (G_OBJECT (contact));
 }
 
 static void
-document_deleted_cb (CouchdbSession *couchdb, const char *dbname, const char *docid, gpointer user_data)
+document_deleted_cb (CouchdbDatabase *database, const char *docid, gpointer user_data)
 {
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (user_data);
 
-	if (g_strcmp0 (dbname, couchdb_backend->dbname) != 0)
-		return;
-
 	e_book_backend_notify_remove (E_BOOK_BACKEND (couchdb_backend), docid);
+
+	/* Remove the contact from the cache */
+	e_book_backend_cache_remove_contact (couchdb_backend->cache, docid);
 }
 
 static GNOME_Evolution_Addressbook_CallStatus
@@ -957,19 +958,20 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 {
 	gchar *uri;
 	const gchar *property;
-	CouchdbDatabaseInfo *db_info;
 	GError *error = NULL;
+	GSList *doc_list, *sl;
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (backend);
 
 	g_return_val_if_fail (E_IS_BOOK_BACKEND_COUCHDB (couchdb_backend), GNOME_Evolution_Addressbook_OtherError);
 
-	if (couchdb_backend->couchdb)
-		g_object_unref (couchdb_backend->couchdb);
-	if (couchdb_backend->dbname)
-		g_free (couchdb_backend->dbname);
+	if (couchdb_backend->couchdb != NULL)
+		g_object_unref (G_OBJECT (couchdb_backend->couchdb));
+	if (couchdb_backend->database != NULL)
+		g_object_unref (G_OBJECT (couchdb_backend->database));
+	if (couchdb_backend->cache != NULL)
+		g_object_unref (G_OBJECT (couchdb_backend->cache));
 
 	/* create CouchDB main object */
-	couchdb_backend->dbname = g_strdup ("contacts");
 	couchdb_backend->using_desktopcouch = FALSE;
 
 	property = e_source_get_property (source, "couchdb_instance");
@@ -988,7 +990,7 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 
 		if (! (couchdb_backend->couchdb = couchdb_session_new (uri))) {
 			g_free (uri);
-			return GNOME_Evolution_Addressbook_OtherError;
+			return GNOME_Evolution_Addressbook_NoSuchBook;
 		}
 
 		g_free (uri);
@@ -996,10 +998,10 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 
 	/* check if only_if_exists */
 	error = NULL;
-	db_info = couchdb_session_get_database_info (couchdb_backend->couchdb,
-						     couchdb_backend->dbname,
-						     &error);
-	if (!db_info) {
+	couchdb_backend->database = couchdb_session_get_database (couchdb_backend->couchdb,
+								  "contacts",
+								  &error);
+	if (couchdb_backend->database == NULL) {
 		if (error) {
 			g_warning ("Could not get CouchDB database info: %s", error->message);
 			g_error_free (error);
@@ -1011,28 +1013,54 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 		/* if it does not exist, create it */
 		error = NULL;
 		if (!couchdb_session_create_database (couchdb_backend->couchdb,
-						      couchdb_backend->dbname,
+						      "contacts",
 						      &error)) {
 			g_warning ("Could not create 'contacts' database: %s", error->message);
 			g_error_free (error);
 
 			return GNOME_Evolution_Addressbook_PermissionDenied;
 		}
-	} else
-		couchdb_database_info_unref (db_info);
+
+		couchdb_backend->database = couchdb_session_get_database (couchdb_backend->couchdb,
+									  "contacts",
+									  &error);
+	}
+
+	/* Create cache */
+	uri = e_source_get_uri (source);
+	couchdb_backend->cache = e_book_backend_cache_new ((const gchar *) uri);
+	g_free (uri);
+
+	/* Populate the cache */
+	e_file_cache_clean (E_FILE_CACHE (couchdb_backend->cache));
+	error = NULL;
+	doc_list = couchdb_database_get_all_documents (couchdb_backend->database,
+						       &error);
+	for (sl = doc_list; sl != NULL; sl = sl->next) {
+		EContact *contact;
+		CouchdbDocument *document = COUCHDB_DOCUMENT (sl->data);
+
+		contact = contact_from_couch_document (document);
+		if (contact != NULL) {
+			e_book_backend_cache_add_contact (couchdb_backend->cache, contact);
+			g_object_unref (G_OBJECT (contact));
+		}
+	}
+
+	couchdb_database_free_document_list (doc_list);
+
+	/* Listen for changes on database */
+	g_signal_connect (G_OBJECT (couchdb_backend->database), "document_created",
+			  G_CALLBACK (document_updated_cb), couchdb_backend);
+	g_signal_connect (G_OBJECT (couchdb_backend->database), "document_updated",
+			  G_CALLBACK (document_updated_cb), couchdb_backend);
+	g_signal_connect (G_OBJECT (couchdb_backend->database), "document_deleted",
+			  G_CALLBACK (document_deleted_cb), couchdb_backend);
+	couchdb_database_listen_for_changes (couchdb_backend->database);
 
 	e_book_backend_set_is_loaded (backend, TRUE);
 	e_book_backend_set_is_writable (backend, TRUE);
 
-	/* Listen for changes on database */
-	g_signal_connect (G_OBJECT (couchdb_backend->couchdb), "document_created",
-			  G_CALLBACK (document_updated_cb), couchdb_backend);
-	g_signal_connect (G_OBJECT (couchdb_backend->couchdb), "document_updated",
-			  G_CALLBACK (document_updated_cb), couchdb_backend);
-	g_signal_connect (G_OBJECT (couchdb_backend->couchdb), "document_deleted",
-			  G_CALLBACK (document_deleted_cb), couchdb_backend);
-	couchdb_session_listen_for_changes (couchdb_backend->couchdb, couchdb_backend->dbname);
-
 	return GNOME_Evolution_Addressbook_Success;
 }
 
@@ -1041,6 +1069,12 @@ e_book_backend_couchdb_remove (EBookBackend *backend, EDataBook *book, guint32 o
 {
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (backend);
 
+	/* Remove the cache */
+	if (couchdb_backend->cache != NULL) {
+		g_object_unref (G_OBJECT (couchdb_backend->cache));
+		couchdb_backend->cache = NULL;
+	}
+
 	/* We don't remove data from CouchDB, since it would affect other apps,
 	   so just report success */
 	e_data_book_respond_remove (book, opid, GNOME_Evolution_Addressbook_Success);
@@ -1057,15 +1091,14 @@ put_document (EBookBackendCouchDB *couchdb_backend, CouchdbDocument *document)
 {
 	GError *error = NULL;
 
-	if (couchdb_document_put (document, couchdb_backend->dbname, &error)) {
+	if (couchdb_database_put_document (couchdb_backend->database, document, &error)) {
 		EContact *new_contact;
-		char *vcard_str;
 
 		/* couchdb_document_put sets the ID for new documents, so need to send that back */
-		vcard_str = vcard_from_couch_document (document);
-		new_contact = e_contact_new_from_vcard (vcard_str);
+		new_contact = contact_from_couch_document (document);
 
-		g_free (vcard_str);
+		/* Add the new contact to the cache */
+		e_book_backend_cache_add_contact (couchdb_backend->cache, new_contact);
 
 		return new_contact;
 	} else {
@@ -1127,14 +1160,14 @@ e_book_backend_couchdb_remove_contacts (EBookBackend *backend,
 		GError *error = NULL;
 		const gchar *uid = (const gchar *) l->data;
 
-		document = couchdb_document_get (couchdb_backend->couchdb, couchdb_backend->dbname, uid, &error);
+		document = couchdb_database_get_document (couchdb_backend->database, uid, &error);
 		if (document) {
 			if (couchdb_backend->using_desktopcouch) {
 				CouchdbStructField *app_annotations, *u1_annotations, *private_annotations;
 
 				/* For desktopcouch, we don't remove contacts, we just
 				 * mark them as deleted */
-				app_annotations = desktopcouch_document_get_application_annotations (document);
+				app_annotations = desktopcouch_document_get_application_annotations (DESKTOPCOUCH_DOCUMENT (document));
 				if (app_annotations == NULL)
 					app_annotations = couchdb_struct_field_new ();
 
@@ -1149,12 +1182,13 @@ e_book_backend_couchdb_remove_contacts (EBookBackend *backend,
 				couchdb_struct_field_set_boolean_field (private_annotations, "deleted", TRUE);
 				couchdb_struct_field_set_struct_field (u1_annotations, "private_application_annotations", private_annotations);
 				couchdb_struct_field_set_struct_field (app_annotations, "Ubuntu One", u1_annotations);
-				desktopcouch_document_set_application_annotations (document, app_annotations);
+				desktopcouch_document_set_application_annotations (DESKTOPCOUCH_DOCUMENT (document), app_annotations);
 
 				/* Now put the new revision of the document */
-				if (couchdb_document_put (document, couchdb_backend->dbname, &error))
+				if (couchdb_database_put_document (couchdb_backend->database, document, &error)) {
 					deleted_ids = g_list_append (deleted_ids, (gpointer) uid);
-				else {
+					e_book_backend_cache_remove_contact (couchdb_backend->cache, uid);
+				} else {
 					if (error != NULL) {
 						g_debug ("Error deleting document: %s", error->message);
 						g_error_free (error);
@@ -1163,13 +1197,14 @@ e_book_backend_couchdb_remove_contacts (EBookBackend *backend,
 				}
 
 				/* Free memory */
-				couchdb_struct_field_unref (app_annotations);
-				couchdb_struct_field_unref (u1_annotations);
-				couchdb_struct_field_unref (private_annotations);
+				g_object_unref (app_annotations);
+				g_object_unref (u1_annotations);
+				g_object_unref (private_annotations);
 			} else {
-				if (couchdb_document_delete (document, &error))
+				if (couchdb_database_delete_document (couchdb_backend->database, document, &error)) {
 					deleted_ids = g_list_append (deleted_ids, (gpointer) uid);
-				else {
+					e_book_backend_cache_remove_contact (couchdb_backend->cache, uid);
+				} else {
 					if (error != NULL) {
 						g_debug ("Error deleting document: %s", error->message);
 						g_error_free (error);
@@ -1237,18 +1272,15 @@ e_book_backend_couchdb_get_contact (EBookBackend *backend,
 				    const char *id)
 {
 	GError *error = NULL;
-	CouchdbDocument *document;
+	EContact *contact;
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (backend);
 
-	document = couchdb_document_get (couchdb_backend->couchdb,
-					 couchdb_backend->dbname,
-					 id,
-					 &error);
-	if (document) {
-		char *vcard = vcard_from_couch_document (document);
+	contact = e_book_backend_cache_get_contact (couchdb_backend->cache, id);
+	if (contact != NULL) {
+		char *vcard = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);
 
-		g_object_unref (G_OBJECT (document));
-		if (vcard) {
+		g_object_unref (G_OBJECT (contact));
+		if (vcard != NULL) {
 			e_data_book_respond_get_contact (book,
 							 opid,
 							 GNOME_Evolution_Addressbook_Success,
@@ -1266,51 +1298,24 @@ e_book_backend_couchdb_get_contact_list (EBookBackend *backend,
 					 EDataBook *book,
 					 guint32 opid, const char *query)
 {
-	GSList *doc_list, *sl;
-	GList *contacts = NULL;
+	GList *doc_list, *contacts = NULL;
 	GError *error = NULL;
-	EBookBackendSExp *card_sexp = NULL;
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (backend);
 
-	card_sexp = e_book_backend_sexp_new (query);
-	if (!card_sexp) {
-		/* XXX this needs to be an invalid query error of some sort*/
-		e_data_book_respond_get_contact_list (book, opid, GNOME_Evolution_Addressbook_OtherError, NULL);
-		return;
-	}
-
-	/* Get the list of documents from CouchDB */
-	doc_list = couchdb_session_list_documents (couchdb_backend->couchdb,
-						   couchdb_backend->dbname,
-						   &error);
-	for (sl = doc_list; sl != NULL; sl = sl->next) {
+	/* Get the list of documents from cache */
+	doc_list = e_book_backend_cache_get_contacts (couchdb_backend->cache, query);
+	while (doc_list != NULL) {
 		char *vcard;
-		CouchdbDocument *document;
-		CouchdbDocumentInfo *doc_info = (CouchdbDocumentInfo *) sl->data;
+		EContact *contact = E_CONTACT (doc_list->data);
 
-		/* Retrieve this document */
-		error = NULL;
-		document = couchdb_document_get (couchdb_backend->couchdb,
-						 couchdb_backend->dbname,
-						 couchdb_document_info_get_docid (doc_info),
-						 &error);
-		if (!document)
-			continue;
+		vcard = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);
+		if (vcard != NULL)
+			contacts = g_list_prepend (contacts, vcard);
 
-		vcard = vcard_from_couch_document (document);
-		if (vcard != NULL) {
-			if (e_book_backend_sexp_match_vcard (card_sexp, vcard))
-				contacts = g_list_prepend (contacts, vcard);
-			else
-				g_free (vcard);
-		}
-
-		g_object_unref (G_OBJECT (document));
+		doc_list = g_list_remove (doc_list, contact);
+		g_object_unref (G_OBJECT (contact));
 	}
 
-	g_object_unref (card_sexp);
-	couchdb_session_free_document_list (doc_list);
-
 	e_data_book_respond_get_contact_list (book, opid, GNOME_Evolution_Addressbook_Success, contacts);
 }
 
@@ -1318,45 +1323,28 @@ static void
 e_book_backend_couchdb_start_book_view (EBookBackend *backend,
 					EDataBookView *book_view)
 {
-	EBookBackendSExp *sexp;
-	GError *error;
-	GSList *doc_list, *sl;
+	GList *doc_list;
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (backend);
 
 	e_book_backend_add_book_view (backend, book_view);
-	sexp = e_data_book_view_get_card_sexp  (book_view);
 
-	/* Get the list of documents from CouchDB */
-	doc_list = couchdb_session_list_documents (couchdb_backend->couchdb,
-						   couchdb_backend->dbname,
-						   &error);
-	for (sl = doc_list; sl != NULL; sl = sl->next) {
-		CouchdbDocument *document;
+	/* Get the list of documents from cache */
+	doc_list = e_book_backend_cache_get_contacts (couchdb_backend->cache,
+						      e_data_book_view_get_card_query (book_view));
+	while (doc_list != NULL) {
 		char *vcard;
-		CouchdbDocumentInfo *doc_info = (CouchdbDocumentInfo *) sl->data;
+		EContact *contact = E_CONTACT (doc_list->data);
 
-		/* Retrieve this document */
-		error = NULL;
-		document = couchdb_document_get (couchdb_backend->couchdb,
-						 couchdb_backend->dbname,
-						 couchdb_document_info_get_docid (doc_info),
-						 &error);
-		if (!document)
-			continue;
-
-		vcard = vcard_from_couch_document (document);
+		vcard = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);
 		if (!vcard)
 			continue;
 
-		if (e_book_backend_sexp_match_vcard (sexp, (const gchar *) vcard))
-			e_data_book_view_notify_update_vcard (book_view, vcard);
-		else
-			g_free (vcard);
+		e_data_book_view_notify_update_vcard (book_view, vcard);
 
-		g_object_unref (G_OBJECT (document));
+		doc_list = g_list_remove (doc_list, contact);
+		g_object_unref (G_OBJECT (contact));
 	}
 
-	couchdb_session_free_document_list (doc_list);
 	e_data_book_view_notify_complete (book_view, GNOME_Evolution_Addressbook_Success);
 }
 
@@ -1416,6 +1404,7 @@ e_book_backend_couchdb_get_supported_fields (EBookBackend *backend,
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_GIVEN_NAME)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_FAMILY_NAME)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_FULL_NAME)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_NAME)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_NICKNAME)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_SPOUSE)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_BIRTH_DATE)));
@@ -1432,7 +1421,6 @@ e_book_backend_couchdb_get_supported_fields (EBookBackend *backend,
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_ORG)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_ORG_UNIT)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_TITLE)));
-	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_ROLE)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_MANAGER)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_ASSISTANT)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_OFFICE)));
@@ -1529,9 +1517,14 @@ e_book_backend_couchdb_dispose (GObject *object)
 		couchdb_backend->couchdb = NULL;
 	}
 
-	if (couchdb_backend->dbname) {
-		g_free (couchdb_backend->dbname);
-		couchdb_backend->dbname = NULL;
+	if (couchdb_backend->cache != NULL) {
+		g_object_unref (G_OBJECT (couchdb_backend->cache));
+		couchdb_backend->cache = NULL;
+	}
+
+	if (couchdb_backend->database) {
+		g_object_unref (G_OBJECT (couchdb_backend->database));
+		couchdb_backend->database = NULL;
 	}
 }
 
@@ -1569,5 +1562,6 @@ static void
 e_book_backend_couchdb_init (EBookBackendCouchDB *backend)
 {
 	backend->couchdb = NULL;
-	backend->dbname = NULL;
+	backend->database = NULL;
+	backend->cache = NULL;
 }
